import requests


class TorrentHandler:
	"""
	Allows to control the Transmission download and upload speed to avoid bandwidth saturation
	"""
	def __init__(self, apiHandler):
		self.apiHandler = apiHandler

	def slowTorrent(self):
		"""
		Send the request to the server to apply alternative speed (slower)
		:return: The request outcome
		"""
		r = requests.post(self.apiHandler, data={'torrent': 'slow_torrent'})
		return r

	def fastTorrent(self):
		"""
		Send the request to the server to restore original Transmission speed
		:return: The request outcome
		"""
		r = requests.post(self.apiHandler, data={'torrent': 'fast_torrent'})
		return r



from .IUBConfiguration import IUBConfiguration
from .FtpUploader import FtpUploader
from .Locker import Locker
from .TorrentHandler import TorrentHandler
from .UploadDb import UploadDb
from .ApiHandler import ApiHandler



import os
import psutil


class Locker:
	"""
	Allows only 1 instance of this software per time
	"""
	def __init__(self, name):
		self.myFile = name
		self.myPid = os.getpid()
		self.processes = self._get_other_processes()

	def _get_other_processes(self):
		processes = []
		for proc in psutil.process_iter(['pid', 'name', 'username']):
			if "python" in proc.info['name'] and self.myPid != proc.info['pid']:
				if self.myFile in proc.cmdline()[1]:
					processes.append(proc)
		return processes

	def already_running(self):
		return len(self.processes) > 0

	def get_processes_same_name(self):
		"""
		Return the list of PID of processes with same name
		:return: The list of PID
		"""
		return [proc.info['pid'] for proc in self.processes]



import yaml


class IUBConfiguration:

	mainSection = 'GlobalSettings'

	def __init__(self, path, logging_handler):
		self.logging = logging_handler
		self.config = self.load_config(path)
		self.logging.getLogger().setLevel(self.config[self.mainSection]['logLevel'])
		self.logging.info('Loaded settings started')

	def load_config(self, path):
		"""
		Load from a Yaml file the settings
		:param path: The path to use to load settings
		:return: The read configuration or stop the script execution
		"""
		with open(path, 'r') as stream:
			try:
				config = yaml.safe_load(stream)
				return config
			except yaml.YAMLError as exc:
				print("Cannot load file: [" + path + "] - Error: " + str(exc))
				self.logging.error("Cannot load file: [" + path + "] - Error: " + str(exc))
				exit()

	def set_config(self, section, name, newValue):
		"""
		Updates a configuration value
		:param section: The name of the section interested by the modify
		:param name: The name of the parameter to update
		:param newValue: The new value to insert
		:return: None
		"""
		if section in self.config and name in self.config[section]:
			self.config[section]['name'] = newValue
		else:
			self.logging.warning(
				"Missing value [" + section + "][" + name + "] from loaded configuration, cannot substitute")

	def get_config(self, section, name=None, fail_silently=True):
		"""
		Retrieve a configuration parameter from loaded settings
		:param section: The name of the section that is requested
		:param name: The name of the parameter inside a section that is requested
		:param fail_silently: True to avoid Exception raise
		:return: The requested configuration value (or set of values)
		"""
		if section not in self.config and name is None:
			# Requested not existent section - Error
			self.logging.warning("Missing section [" + section + "] from loaded configuration, checking in main section")
			if fail_silently:
				return None
			else:
				raise Exception("Missing required section [" + section + "] from loaded configuration")
		elif section in self.config and name is None:
			# Requested existent section - Return entire section
			self.logging.info("Retrieving entire section: [" + section + "]")
			return self.coalesce_section(self.config[section])
		else:
			# Requested specific value
			if name not in self.coalesce_section(self.config[section]):
				self.logging.warning("Missing value [" + name + "] from section [" + section + "]")
				if fail_silently:
					return None
				else:
					raise Exception("Missing required value [" + name + "] from section [" + section + "]")
			else:
				self.logging.debug("Retrieving value [" + name + "] from section: [" + section + "]")
				return self.coalesce_section(self.config[section])[name]

	def coalesce_section(self, first_section, second_section=None):
		"""
		Join the first passed section with the second one. If missing the second section, join with main section
		:param first_section: The first section to join, all parameters and values in this section will be kept
		:param second_section: The second section to join, in case of value already existing in the first one this
								value will not be considered
		:return: The joined values
		"""
		if second_section is None:
			second_section = self.config[self.mainSection]
		return {**second_section, **first_section}



import requests
import json
import time
import random


class ApiHandler:
	"""
	A class that manages the IUB api
	username: The IUB username
	apiSite: The IUB api endpoint
	tokenPath: The path to an empty file that contains the IUB API token
	loggingHandler: The logging handler where any error is sent
	"""
	def __init__(self, username, apiSite, tokenPath, loggingHandler):
		self.username = username
		self.apiSite = apiSite
		self.tokenPath = tokenPath
		self.logging = loggingHandler
		#Read token
		self.token = self.readToken()
		self.toRestore = {}
	
	res = {
		"ONLINE": "Files already online",
		"WAIT": "Wait before next refresh",
		"RESTORING": True,
		"ALREADY_REQ": "Link already requested"
	}

	def readToken(self):
		"""
		Read the token from the given url
		:return: The read API token
		"""
		file = open(self.tokenPath, "r")
		if not file:
			self.logging.error('Token file not read - Cannot execute any request')
			exit()
		token = file.read().strip()
		self.logging.info('ApiHandler - readToken - IUB Token read from file '+self.tokenPath)
		return token

	def getActiveGenres(self):
		"""
		Request the available genres
		:return: The list of active genres
		"""
		req = "getActiveGenre"
		r = requests.post(self.apiSite+"/api/genre.php", data={'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)

	def searchRelease(self, title):
		"""
		Search a certain title
		:param title: The title to search
		:return:
		"""
		req = "searchTitles"
		r = requests.post(self.apiSite+"/api/release.php", data={'user': self.username, 'psw': self.token, 'req': req, 'title': title})
		return self.getParsedResponse(r)
	
	#Download a certain release
	def downloadRelease(self, releaseId):
		req = "download_tm"
		r = requests.post(self.apiSite+"/api/download.php", data={'user': self.username, 'psw': self.token, 'req': req,  'code': releaseId})
		return self.getParsedResponse(r)
	
	#Array of IUB genres 
	# 0	-> id
	# 1	-> name
	def getAllGenres(self):
		req = "listAllGenre"
		r = requests.post(self.apiSite+"/api/genre.php", data={'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
	
	#Return the list of all releases to save
	def getAllReleasesToSave(self):
		self.logging.info('Requested all releases that need to be saved')
		req = "get_all_releases_to_save"
		r = requests.post(self.apiSite+"/api/release_saver.php", data={'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
	
	#Retrieve the list of all releases in a free 1fichier account from the server		
	def getAllReleasesPerFreeAccount(self):
		self.logging.info('Requested all releases with a free account')
		req = "get_all_releases_per_account"
		r = requests.post(self.apiSite+"/api/release.php", data={'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
	
	#Retrieve the list of all materials having the given materials
	def getAllReleases(self, genres):
		self.logging.info("Request all releases present in the server")
		req = "get_all_releases"
		r = requests.post(self.apiSite+"/api/release.php", data={'user': self.username, 'psw': self.token, 'req': req, 'genres': json.dumps(genres)})
		return self.getParsedResponse(r)
		
	#Creates the dictionaries
	def manageReleases(self):
		all_rel = self.getAllReleasesPerFreeAccount()
		#Creates a list of all releases ignoring the genre
		for genre in all_rel:
			for account in all_rel[genre]:
				#Creates array in the dictionary if needed
				try:
					self.toRestore[account]
				except KeyError:
					self.logging.info("Generated key in library for: "+account)
					self.toRestore[account] = []
				
				#Populate dictionary
				for release in all_rel[genre][account]:
					self.toRestore[account].append(release)
		
		#Log the account to check
		for account in self.toRestore:
			now = len(self.toRestore[account])
			self.logging.info("Account: "+account+" - To check "+str(now)+" releases")
	
	#Restore a single release
	def restoreRelease(self, code):
		self.logging.debug('Request restore: '+str(code))
		req = "refresh_1f"
		r = requests.post(self.apiSite+"/api/release_refresher.php", data={'code': code, 'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
		
	#Start restoring all until I can only wait
	def restoreAll(self):
		cloneToRestore = self.toRestore.copy()
		#Iterate over all account
		for account in cloneToRestore:
			now = len(self.toRestore[account])
			print("START: Account: "+account+" - To check "+str(now)+" releases")
			#Create initial copy and iterate over it
			new_list = self.toRestore[account].copy()
			#Iterate over all the releases
			for code in new_list:
				try:
					resp = self.restoreRelease(code)

					if resp == self.res["ONLINE"]:
						self.logging.debug('Already online: '+str(code))
						self.removeObject(account, code)
					elif resp == self.res["WAIT"]:
						#Exit from this loop and not remove the object!
						self.logging.debug('Wait before new request: '+str(code)+'\nExit from loop')
						break
					elif self.res["RESTORING"]:
						print("Restoring: "+str(code))
						self.logging.info('Restore request successful: '+str(code))
						self.removeObject(account, code)
					elif resp == self.res["ALREADY_REQ"]:
						print("Already requested: "+str(code))
						self.logging.warning('Unexpected: Restore already requested of: '+str(code))
						self.removeObject(account, code)
				except Exception as e:
					print("Problem restoring release: " + str(code) + " [" + str(e) + "]")
					self.logging.error("Problem restoring release: "+str(code))
				
				print(str(random.randrange(0, 1000)/1000))
				time.sleep(random.randrange(0, 1000)/1000)
				
			after = len(self.toRestore[account])
			print("STOP: Account: "+account+" - To check "+str(after)+" releases")
			#Remove empty dictionaries
			if not after:
				lib_bef = len(self.toRestore)
				print(account+" removed")
				self.toRestore.pop(account, None)
				lib_aft = len(self.toRestore)
				self.logging.info("Removed account: "+account+" - Before: "+str(lib_bef)+' elem - After: '+str(lib_aft))
			else:
				print("There are still "+str(after)+" elements for account: "+account)

	def removeObject(self, account, item):
		"""
		Remove an object from the original library
		:param account: The account used for the item
		:param item: The item to remove
		:return:
		"""
		self.logging.debug('Remove: '+str(item)+' from '+account)
		elem_before = len(self.toRestore[account])
		self.toRestore[account].remove(item)
		elem_after = len(self.toRestore[account])
		self.logging.debug('-->Before: '+str(elem_before)+' Now: '+str(elem_after))

	def countObjectLeft(self):
		"""
		Return the count of objects left
		:return: The number of objects left
		"""
		numElem = 0
		for account in self.toRestore:
			numElem += len(self.toRestore[account])
		return numElem
	
	#Insert a new material		
	def insertNewMaterial(self, type_file, number_files):
		self.logging.info('Inserting: '+type_file)
		req = "insert_new_material"
		#Send request
		r = requests.post(
			self.apiSite+"/api/release.php",
			data={
				'user': self.username, 
				'psw': self.token, 
				'req': req, 
				'num': number_files,
				'material': type_file
			}
		)
		#Control if the response is ok
		try:
			res = self.getParsedResponse(r)
			print("Parsed json response: "+str(res))
			if "upped" in res:
				return int(res["upped"])
			else:
				self.logging.error("Not found appropriate upped key ["+r.text+"]")
				return 0
		except (ValueError, TypeError) as e:
			self.logging.error("Error decoding Json response ["+r.text+"] [" + str(e) + "]")
			return 0

	#Return the count of left objects
	def orderThisRelease(self, code):
		self.logging.debug('Ordering: '+str(code))
		req = "order_prem_dir_fichier"
		r = requests.post(self.apiSite+"/api/release.php", data={'code': code, 'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
	
	#Request the available genres
	def saveRelease(self, code):
		req = "save_release"
		r = requests.post(self.apiSite+"/api/release_saver.php", data={'user': self.username, 'psw': self.token, 'req': req, 'code': code})
		return self.getParsedResponse(r)

	#Request to refresh the premium links inside the DB taking them from the directory
	def refreshPremiumLinks(self, code):
		req = "new_1f_links"
		r = requests.post(self.apiSite+"/api/release.php", data={'user': self.username, 'psw': self.token, 'req': req, 'code': code})
		return self.getParsedResponse(r)
	
	#Refresh torrent cache
	def refreshTorrentCache(self):
		req = "icv_refresh_cache"
		r = requests.post(self.apiSite+"/api/torrent.php", data={'user': self.username, 'psw': self.token, 'req': req})
		return self.getParsedResponse(r)
			
	def decode(self, r):
		try:
			return json.loads(r.text)
		except ValueError as e:
			self.logging.error("Cannot decode: " + r.text + "[" + str(e) + "]")
			return r.text
		except Exception as e:
			self.logging.exception("JSON decode error [" + str(e) + "]")

	def getParsedResponse(self, result):
		result = self.decode(result)
		if type(result) is not dict:
			self.logging.error("Expected a dict, obtained a " + type(result))
			print("Expected a dict, obtained a " + type(result))
			return result
		else:
			return result['result']


class UnknownTypeFile(Exception):
	"""
	Custom exceptions for unknown file type
	"""
	pass



import ftplib
import os
import time
import sys


class FtpUploader:
	"""
	Allows to upload to a FTP account
	"""
	
	def __init__(self, host, username, passwordFile, port=21, logging=None):
		if logging:
			self.logging = logging
		#Retrieve password
		with open(passwordFile) as fp:
			password = fp.readline().strip()

		#Enstablish connection
		self.connection = ftplib.FTP_TLS()
		self.connection.connect(host, port)
		if self.logging:
			self.logging.info("FtpUploader - uploadFile - Connected to ["+host+"]")
		self.connection.login(username, password)
		if self.logging:
			self.logging.info("FtpUploader - uploadFile - Logged as ["+username+"]")
		self.connection.prot_p()
		if self.logging:
			self.logging.info("FtpUploader - uploadFile - Connection secured")

		self.writtenSize = 0
		self.maxSpeed = 0
		self.uploadStart = None

	def addLogging(self, logging):
		"""
		Assign a logging handler
		"""
		self.logging = logging
		if logging:
			self.logging.info("FtpUploader - Assigned logging handler")
		else:
			print("No logging handler given")

	def uploadFile(self, file_path, block_size=262144, max_speed=0):
		"""
		Upload a file - Return the upload status
		block_size	The size of the block to send - Default 256 KiB
		maxSpeed	Speed in KB/s to maintain during transfer - Default 0 -> No limit
		"""
		self.maxSpeed = max_speed * 1024
		if not self.maxSpeed:
			print("Transferring at maximum speed")
		self.uploadStart = time.time()
		name = os.path.basename(file_path).encode("utf-8", "ignore").decode("latin-1", "ignore")
		with open(file_path, 'rb') as fp:
			try:
				self.connection.storbinary('STOR ' + name, fp, block_size, self.throttler)
				if self.logging:
					self.logging.info("FtpUploader - uploadFile - Uploaded [" + name + "] - [" + file_path + "]")
					print("FtpUploader - uploadFile - Uploaded [" + name + "] - [" + file_path + "]")
				return True
			except ftplib.error_perm:
				self.logging.warn("FtpUploader - uploadFile - Permanent error during upload [" + name + "] - [" + file_path + "]")
			except ftplib.error_temp:
				self.logging.warn("FtpUploader - uploadFile - Temporary error during upload [" + name + "] - [" + file_path + "]")
		return False
	
	total_length = 0
	start_time = time.time()

	#This function slow down the transmission during
	def throttler(self, buf):
		i = 0
		sleepTime = 0.1
		sleepSeconds = 5
		#Do nothing if no limit set to max speed
		if not self.maxSpeed:
			return
		#Get the written bytes
		self.writtenSize += sys.getsizeof(buf)
		cycles = sleepSeconds/sleepTime
		while self.writtenSize / (time.time() - self.uploadStart) > self.maxSpeed:
			time.sleep(sleepTime)
			i += 1
			if not i % cycles:
				print("Elapsed: "+str(time.time() - self.uploadStart)+"s, started at "+str(self.uploadStart)+" and written "+str(self.writtenSize/1024/1024)+" MB Sleeping since "+str(sleepSeconds)+" seconds")
				self.logging.info("Elapsed: "+str(time.time() - self.uploadStart)+"s, started at "+str(self.uploadStart)+" and written "+str(self.writtenSize/1024/1024)+" MB Sleeping since "+str(sleepSeconds)+" seconds")
	
	#Tries to gracefully exit
	def __del__(self):
		self.connection.quit()
		self.logging.info("FtpUploader - __del__ - Quit gracefully")




class UploadDb:
	def __init__(self, path):
		self.path = path

	def saveDataPlain(self, dataset):
		"""
		Store an array into the DB
		:param dataset: List of entry data to store in the DB
		:return: None
		"""
		with open(self.path, 'w') as fd:
			for item in dataset:
				fd.write(item.strip()+"\n")
		fd.close()

	def appendDataPlain(self, line):
		"""
		Append a single line
		:param line: The new record to add to DB
		:return: None
		"""
		fd = open(self.path, 'a')
		fd.write(line+"\n")
		fd.close()

	def restoreDataPlain(self):
		"""
		Restore all the data as an array
		:return: The list of record read from DB
		"""
		fd = open(self.path, 'r')
		content = fd.readlines()
		val = [x.strip() for x in content]
		fd.close()
		return val
